'''
Utilities for scripts
'''

import sys
from six import iteritems

__all__ = ['parse_args']

def parse_args(num_args, allowed_params, msg, string_args=set()):
    msg += '\nAllowed arguments and default values:\n'
    for k, v in iteritems(allowed_params):
        msg += '\n    %s = %s' % (k, v)
    if len(sys.argv)<=num_args:
        print(msg)
        exit(1)
    params = {}
    for spec in sys.argv[num_args+1:]:
        name, val = spec.split('=')
        if name not in string_args:
            if val.lower()=='true':
                val = 'True'
            elif val.lower()=='false':
                val = 'False'
            val = eval(val)
        params[name] = val
    for k in list(params.keys()):
        if k not in allowed_params:
            print(msg)
            exit(1)
    for k, v in iteritems(allowed_params):
        if k not in params:
            params[k] = v
    return sys.argv[1:num_args+1], params



#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True
# distutils: language = c++

import numpy
cimport numpy
from numpy.random import randint

from cython cimport integral, floating

from libcpp.vector cimport vector

cdef integral mask_difference(integral[:] unmasked, integral start1, integral end1, integral start2, integral end2):
    cdef integral i1, i2, u1, u2, d
    cdef integral n1 = end1-start1
    cdef integral n2 = end2-start2
    i1 = start1
    i2 = start2
    d = 0
    while i1<end1 or i2<end2:
        if i1>=end1:
            # there now cannot be a match to u2, so we increase distance by 1
            d += 1
            i2 += 1
            continue
        if i2>=end2:
            # there now cannot be a match to u1, so we increase distance by 1
            d += 1
            i1 += 1
            continue
        u1 = unmasked[i1]
        u2 = unmasked[i2]
        if u1==u2:
            # we match u1 and u2 so increment both pointers to put them out of the pool and since they're matched
            # we don't increase the distance
            i1 += 1
            i2 += 1
        elif u1<u2:
            # we remove u1 from the pool and it didn't match, so we increase the distance
            d += 1
            i1 += 1
        elif u2<u1:
            # we remove u2 from the pool and it didn't match, so we increase the distance
            d += 1
            i2 += 1
    return d


cpdef do_mask_starts(integral[:] clusters,
                     integral[:] unmasked,
                     integral[:] ustart,
                     integral[:] uend,
                     integral[:] allspikes,
                     integral num_features,
                     integral num_clusters,
                     ):
    cdef vector[integral] best_ids
    cdef vector[integral] candidate_ids
    cdef vector[integral] candidate_ends
    cdef integral cur_cluster = 0
    cdef integral p, mask_id, best_distance, candidate_id, candidate_end, c_idx, d, p_idx
    found = dict()
    end = dict()
    for p_idx in range(len(allspikes)):
        p = allspikes[p_idx]
        mask_id = ustart[p]
        if mask_id in found:
            # we've already used this mask
            clusters[p] = found[mask_id]
        else:
            if cur_cluster<num_clusters:
                # use a new mask
                found[mask_id] = cur_cluster
                end[mask_id] = uend[p]
                clusters[p] = cur_cluster
                cur_cluster += 1
                candidate_ids.push_back(mask_id)
                candidate_ends.push_back(uend[p])
            else:
                # we have to find the closest mask (this is the computationally intensive bit!)
                best_distance = num_features+1
                best_ids.clear()
                for c_idx in range(candidate_ids.size()):
                    candidate_id = candidate_ids[c_idx]
                    candidate_end = candidate_ends[c_idx]
                    d = mask_difference(unmasked, ustart[p], uend[p], candidate_id, candidate_end)
                    if d==best_distance:
                        best_ids.push_back(candidate_id)
                    elif d<best_distance:
                        best_distance = d
                        best_ids.clear()
                        best_ids.push_back(candidate_id)
                best_id = best_ids[randint(best_ids.size())]
                clusters[p] = found[best_id]
                found[mask_id] = found[best_id]
                end[mask_id] = end[best_id]



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand
from six.moves import range

# we use the version that is used in klustakwik2 rather than separately testing the numba/cython
# versions
from klustakwik2.clustering import accumulate_cluster_mask_sum

from .test_io import generate_simple_test_raw_data

@nottest
def generate_simple_test_kk(**params):
    raw_data, fet, fmask, features, correction_terms = generate_simple_test_raw_data()
    data = raw_data.to_sparse_data()
    kk = KK(data, **params)
    kk.orig_fet = fet
    kk.orig_fmask = fmask
    kk.orig_features = features
    kk.orig_correction_terms = correction_terms
    clusters = array([2, 3, 3, 4])
    kk.initialise_clusters(clusters)
    return kk

def test_accumulate_cluster_mask_sum():
    kk = generate_simple_test_kk()
    fet = kk.orig_fet
    fmask = kk.orig_fmask
    
    assert kk.num_clusters_alive==5

    num_clusters = kk.num_clusters_alive
    num_features = kk.num_features
    
    cluster_mask_sum = zeros((num_clusters, num_features))
    cluster_mask_sum[:2, :] = -1 # ensure that clusters 0 and 1 are masked

    for cluster in range(2, num_clusters):
        accumulate_cluster_mask_sum(kk, cluster_mask_sum[cluster, :], kk.get_spikes_in_cluster(cluster))
    
    # cluster 2 has only point 0 in it, so the cluster_mask sum should be just the corresponding
    # fmask line
    assert_array_almost_equal(cluster_mask_sum[2, :], fmask[0, :])
    # similarly for the others
    assert_array_almost_equal(cluster_mask_sum[3, :], fmask[1, :]+fmask[2, :])
    assert_array_almost_equal(cluster_mask_sum[4, :], fmask[3, :])
    assert (cluster_mask_sum[0, :]==-1).all()
    assert (cluster_mask_sum[1, :]==-1).all()


if __name__=='__main__':
    test_accumulate_cluster_mask_sum()



'''
Default clustering parameters
'''

from numpy import log

default_parameters = dict(
     prior_point=1,
     mua_point=2,
     noise_point=1,
     points_for_cluster_mask=100,
     penalty_k=0.0,
     penalty_k_log_n=1.0,
     max_iterations=1000,
     num_starting_clusters=500,
     use_noise_cluster=True,
     use_mua_cluster=True,
     num_changed_threshold=0.05,
     full_step_every=1,
     split_first=20,
     split_every=40,
     max_possible_clusters=1000,
     dist_thresh=log(10000.0),
     max_quick_step_candidates=100000000, # this uses around 760 MB RAM
     max_quick_step_candidates_fraction=0.4,
     always_split_bimodal=False,
     subset_break_fraction=0.01,
     break_fraction=0.0,
     fast_split=False,
     max_split_iterations=None,
     consider_cluster_deletion=True,
     num_cpus=None,
     )



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from klustakwik2.linear_algebra import BlockPlusDiagonalMatrix
from scipy.linalg import cho_factor, cho_solve, cholesky, solve_triangular
from klustakwik2.numerics.cylib.e_step import trisolve
from numpy.random import randn
from numpy.linalg import inv
from six.moves import range

def test_cholesky_trisolve():
    M = array([[3.5, 0,   0,   0],
               [0,   1.2, 0,   0.1],
               [0,   0,   6.2, 0],
               [0,   0.1,   0,   11]])
    masked = array([0, 2])
    unmasked = array([1, 3])
    cov = BlockPlusDiagonalMatrix(masked, unmasked)
    cov.block[:, :] = M[ix_(unmasked, unmasked)]
    cov.diagonal[:] = M[masked, masked]
    # test cholesky decomposition
    chol = cov.cholesky()
    M_chol_from_block = zeros_like(M)
    M_chol_from_block[ix_(unmasked, unmasked)] = chol.block
    M_chol_from_block[masked, masked] = chol.diagonal
    M_chol = cholesky(M, lower=True)
    assert_array_almost_equal(M_chol_from_block, M_chol)
    assert_array_almost_equal(M, M_chol.dot(M_chol.T))
    assert_array_almost_equal(cov.block, chol.block.dot(chol.block.T))
    # test trisolve
    x = randn(M.shape[0])
    y1 = solve_triangular(M_chol, x, lower=True)
    y2 = chol.trisolve(x)
    y3 = zeros(len(x))
    trisolve(chol.block, chol.diagonal, chol.masked, chol.unmasked, len(chol.masked), len(chol.unmasked), x, y3)
    assert_array_almost_equal(y1, y2)
    assert_array_almost_equal(y1, y3)
    assert_array_almost_equal(y2, y3)
    # test compute diagonal of inverse of cov matrix used in E-step
    inv_cov_diag = zeros(len(x))
    basis_vector = zeros(len(x))
    for i in range(len(x)):
        basis_vector[i] = 1.0
        root = chol.trisolve(basis_vector)
        inv_cov_diag[i] = sum(root**2)
        basis_vector[:] = 0
    M_inv_diag = diag(inv(M))
    assert_array_almost_equal(M_inv_diag, inv_cov_diag)


if __name__=='__main__':
    test_cholesky_trisolve()




'''
Test clustering performance on synthetic dataset.
'''

from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand, randn
from six.moves import range
import unittest

@nottest
def test_approximately_well_clustered(clusters, num_clusters, cluster_size, fraction=0.02):
    b = bincount(clusters)
    assert b[0]<=fraction*cluster_size
    assert b[1]<=fraction*cluster_size
    dominant_clusters = []
    for i in range(num_clusters):
        cur_clusters = clusters[i*cluster_size:(i+1)*cluster_size]
        b = bincount(cur_clusters)
        dc = argmax(b)
        dominant_clusters.append(dc)
        assert cluster_size-b[dc]<=fraction*cluster_size
    assert len(dominant_clusters)==len(unique(dominant_clusters)) 
    

@nottest
def generate_synthetic_data(num_features, spikes_per_centre, centres, save_to_fet=None):
    '''
    Generates data that comes from a distribution with multiple centres. centres is a list of
    tuples (fet_mean, fet_std, fmask_mean, fmask_std) and features and masks are generated via
    normal random numbers (fmask is clipped between 0 and 1).

    TODO: data generated by this isn't normalised in [0,1]
    '''
    fet = []
    fmask = []
    offsets = []
    unmasked = []
    n = 0
    num_spikes = len(centres)*spikes_per_centre
    offsets = [n]
    fsum = zeros(num_features)
    fsum2 = zeros(num_features)
    nsum = zeros(num_features)
    if save_to_fet is not None:
        fetfile = open(save_to_fet+'.fet.1', 'wt')
        fmaskfile = open(save_to_fet+'.fmask.1', 'wt')
        fetfile.write('%d\n' % num_features)
        fmaskfile.write('%d\n' % num_features)
    for c, s, fmc, fms in centres:
        c = array(c, dtype=float)
        s = array(s, dtype=float)
        fmc = array(fmc, dtype=float)
        fms = array(fms, dtype=float)
        for i in range(spikes_per_centre):
            f = randn(num_features)*s+c
            fm = clip(randn(num_features)*fms+fmc, 0, 1)
            if save_to_fet is not None:
                fetfile.write(' '.join(map(str, f))+'\n')
                fmaskfile.write(' '.join(map(str, fm))+'\n')
            u, = (fm>0).nonzero()
            m, = (fm==0).nonzero()
            u = array(u, dtype=int)
            fet.append(f[u])
            fmask.append(fm[u])
            unmasked.append(u)
            n += len(u)
            offsets.append(n)
            fsum[m] += f[m]
            fsum2[m] += f[m]**2
            nsum[m] += 1
    fet = hstack(fet)
    fmask = hstack(fmask)
    unmasked = hstack(unmasked)
    offsets = array(offsets)
    return RawSparseData(fsum/nsum, fsum2/nsum-(fsum/nsum)**2,
                         fet, fmask, unmasked, offsets).to_sparse_data()


def test_synthetic_2d_trivial():
    '''
    This is the most trivial clustering problem, two well separated clusters in 2D with almost
    no noise and perfect starting masks. All the algorithm has to do is not do anything. We
    therefore test that it gives perfect results.
    '''
    data = generate_synthetic_data(2, 100, [
       # fet mean, fet var,      fmask mean, fmask var
        ((1, 0),   (0.01, 0.01), (1, 0),     (0.0, 0.0)),
        ((0, 1),   (0.01, 0.01), (0, 1),     (0.0, 0.0)),
        ])
    kk = KK(data, points_for_cluster_mask=1e-100, num_starting_clusters=10)
    kk.cluster_mask_starts()
    test_approximately_well_clustered(kk.clusters, 2, 100)


def test_synthetic_2d_easy():
    '''
    In this test, there is a small amount of variance in both the masks and the features.
    Note that if you put the average mask at 1 instead of 0.5 it fails because the 'corrected'
    data will be non-Gaussian (as it gets clipped between 0 and 1). Similarly, if you put the noise
    in the data as 0 it doesn't work because you get singular matrices.
    '''
    data = generate_synthetic_data(2, 100, [
        ((1, 0), (0.01,)*2, (0.5, 0), (0.01, 0.0)),
        ((0, 1), (0.01,)*2, (0, 0.5), (0.0, 0.01)),
        ])
    kk = KK(data, points_for_cluster_mask=1e-100, num_starting_clusters=10)
    kk.cluster_mask_starts()
    test_approximately_well_clustered(kk.clusters, 2, 100)


def test_synthetic_4d_easy():
    data = generate_synthetic_data(4, 1000, [
        ((1, 1, 0, 0), (0.1,)*4, (1.5, 0.5, 0, 0), (0.05, 0.05, 0, 0)),
        ((0, 1, 1, 0), (0.1,)*4, (0, 0.5, 1.5, 0), (0, 0.05, 0.05, 0)),
        ((0, 0, 1, 1), (0.1,)*4, (0, 0, 0.5, 1.5), (0, 0, 0.05, 0.05)),
        ((1, 0, 0, 1), (0.1,)*4, (1.5, 0, 0, 1.5), (0.05, 0, 0, 0.05)),
        ])
    kk = KK(data, points_for_cluster_mask=1e-100, num_starting_clusters=20)
    kk.cluster_mask_starts()
    test_approximately_well_clustered(kk.clusters, 4, 1000)


def test_synthetic_4d_trivial():
    data = generate_synthetic_data(4, 1000, [
        ((1, 1, 0, 0), (0.1,)*4, (1.5, 0.5, 0, 0), (0,)*4),
        ((0, 1, 1, 0), (0.1,)*4, (0, 0.5, 1.5, 0), (0,)*4),
        ((0, 0, 1, 1), (0.1,)*4, (0, 0, 0.5, 1.5), (0,)*4),
        ((1, 0, 0, 1), (0.1,)*4, (1.5, 0, 0, 1.5), (0,)*4),
        ])
    kk = KK(data, points_for_cluster_mask=1e-100, num_starting_clusters=20)
    kk.cluster_mask_starts()
    test_approximately_well_clustered(kk.clusters, 4, 1000)


@unittest.skip("Tmporarily disabled until we work out why cycles are happening here")
def test_synthetic_4d_easy_non_gaussian():
    data = generate_synthetic_data(4, 1000, [
        ((1, 1, 0, 0), (0.1,)*4, (1.5, 0.5, 0, 0), (0.05, 0.05, 0.01, 0)),
        ((0, 1, 1, 0), (0.1,)*4, (0, 0.5, 1.5, 0), (0, 0.05, 0.05, 0.01)),
        ((0, 0, 1, 1), (0.1,)*4, (0, 0, 0.5, 1.5), (0.01, 0, 0.05, 0.05)),
        ((1, 0, 0, 1), (0.1,)*4, (1.5, 0, 0, 1.5), (0.05, 0, 0.01, 0.05)),
        ])
    # no space for error, so we set quick steps off
    kk = KK(data, full_step_every=1, points_for_cluster_mask=1e-100, num_starting_clusters=20)
    kk.cluster_mask_starts()
    test_approximately_well_clustered(kk.clusters, 4, 1000)

def test_splitting():
    data = generate_synthetic_data(4, 1000, [
        ((1, 1, 0, 0), (0.1,)*4, (1.5, 0.5, 0, 0), (0.05, 0.05, 0, 0)),
        ((0, 1, 1, 0), (0.1,)*4, (0, 0.5, 1.5, 0), (0, 0.05, 0.05, 0)),
        ((0, 0, 1, 1), (0.1,)*4, (0, 0, 0.5, 1.5), (0, 0, 0.05, 0.05)),
        ((1, 0, 0, 1), (0.1,)*4, (1.5, 0, 0, 1.5), (0.05, 0, 0, 0.05)),
        ])
    kk = KK(data, points_for_cluster_mask=1e-100, num_starting_clusters=20,
            split_every=1)
    kk.cluster_mask_starts()


if __name__=='__main__':
#     console_log_level('debug')
    for _ in range(100):
        test_synthetic_2d_trivial()
        test_synthetic_2d_easy()
        test_synthetic_4d_easy()
        test_synthetic_4d_trivial()
        test_synthetic_4d_easy_non_gaussian()
        test_splitting()


from numpy import *
from .e_step_cy import *
import multiprocessing
from six import itervalues

__all__ = ['compute_log_p_and_assign']

def compute_log_p_and_assign(kk, cluster, weight, inv_cov_diag, log_root_det, chol, cluster_mean,
                             only_evaluate_current_clusters, n_cpu=None):
    num_clusters = len(kk.num_cluster_members)
    num_features = kk.num_features
    num_spikes = kk.num_spikes

    log_addition = log_root_det-log(weight)+0.5*log(2*pi)*num_features

    data = kk.data
    unmasked = data.unmasked
    ustart = data.unmasked_start
    uend = data.unmasked_end
    features = data.features
    vstart = data.values_start
    vend = data.values_end


    log_p_best = kk.log_p_best
    log_p_second_best = kk.log_p_second_best
    noise_mean = data.noise_mean
    noise_variance = data.noise_variance
    correction_terms = data.correction_terms

    clusters = kk.clusters
    clusters_second_best = kk.clusters_second_best
    old_clusters = kk.old_clusters
    full_step = kk.full_step

    if n_cpu is None:
        n_cpu = multiprocessing.cpu_count()
    f2cm_multiple = numpy.zeros(num_features*n_cpu)
    root_multiple = numpy.zeros(num_features*n_cpu)
    cluster_log_p = numpy.zeros(num_spikes)

    if only_evaluate_current_clusters:
        candidates = kk.quick_step_candidates[cluster]
        num_spikes = len(candidates)
    elif full_step:
        candidates = zeros(0, dtype=int)
    else:
        candidates = kk.quick_step_candidates[cluster]
        num_spikes = len(candidates)

    do_log_p_assign_computations(
                                  noise_mean, noise_variance, cluster_mean, correction_terms,
                                  log_p_best, log_p_second_best,
                                  clusters, clusters_second_best, old_clusters,
                                  full_step,
                                  inv_cov_diag,
                                  unmasked, ustart, uend, features, vstart, vend,
                                  root_multiple, f2cm_multiple,
                                  num_features, num_spikes, log_addition, cluster,
                                  chol.block, chol.diagonal, chol.masked, chol.unmasked,
                                  n_cpu,
                                  cluster_log_p,
                                  candidates,
                                  only_evaluate_current_clusters,
                                  )

    if only_evaluate_current_clusters:
        kk.log_p_best[candidates] = cluster_log_p[candidates]
    elif full_step and kk.collect_candidates and hasattr(kk, 'old_log_p_best') and kk.full_step_every>1:
        max_quick_step_candidates = min(kk.max_quick_step_candidates,
            kk.max_quick_step_candidates_fraction*kk.num_spikes*kk.num_clusters_alive)
        candidates, = (cluster_log_p-kk.old_log_p_best<=kk.dist_thresh).nonzero()
        kk.quick_step_candidates[cluster] = array(candidates, dtype=int)
        num_candidates = sum(len(v) for v in itervalues(kk.quick_step_candidates))
        if num_candidates>max_quick_step_candidates:
            kk.collect_candidates = False
            kk.quick_step_candidates.clear()
            kk.force_next_step_full = True
            if num_candidates>kk.max_quick_step_candidates:
                kk.log('info', 'Ran out of storage space for quick step, try increasing '
                               'max_quick_step_candidates if this happens often.')
            else:
                kk.log('debug', 'Exceeded quick step point fraction, next step '
                                'will be full')

    return kk.num_spikes-num_spikes



#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True

import numpy
cimport numpy

from cython cimport integral, floating

cpdef doaccum(integral[:] spikes,
              integral[:] unmasked,
              integral[:] ustart,
              integral[:] uend,
              floating[:] masks,
              integral[:] vstart,
              integral[:] vend,
              floating[:] cluster_mask_sum,
             ):
    cdef integral pp, p, c, num_unmasked, i, j, k
    for pp in range(len(spikes)):
        p = spikes[pp]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            j = unmasked[ustart[p]+i]
            k = vstart[p]+i
            cluster_mask_sum[j] += masks[k]



from .data import *
from .input_output import *
from .mask_starts import *
from .clustering import *
from .logger import *
from .scripts.tools import *
from .monitoring import *
from .debugtools import *
from .default_parameters import *

__version__ = '0.2.4'
__version__ = get_kk_version(__version__)

log_message('info', 'KlustaKwik2 version '+__version__)



#cython: boundscheck=False
#cython: wraparound=False
#cython: cdivision=True
#cython: infer_types=True

import numpy
cimport numpy

from cython cimport integral, floating

cpdef do_compute_cluster_mean(
             integral[:] spikes,
             integral[:] unmasked,
             integral[:] ustart,
             integral[:] uend,
             floating[:] features,
             integral[:] vstart,
             integral[:] vend,
             floating[:] cluster_mean,
             integral[:] num_added,
             floating[:] noise_mean,
             floating prior,
             ):
    cdef integral pp, p, c, j, k, num_unmasked
    cdef integral num_features = cluster_mean.shape[0]
    cdef integral num_cluster_members
    num_cluster_members = len(spikes)
    for pp in range(num_cluster_members):
        p = spikes[pp]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            j = unmasked[ustart[p]+i]
            k = vstart[p]+i
            cluster_mean[j] += features[k]
            num_added[j] += 1
    if num_cluster_members==0:
        return
    for i in range(num_features):
        cluster_mean[i] += noise_mean[i]*(num_cluster_members-num_added[i]) 
        cluster_mean[i] += prior*noise_mean[i]
        cluster_mean[i] /= num_cluster_members+prior


cpdef do_var_accum(
            integral[:] spike_indices,
            floating[:] cluster_mean,
            floating[:] noise_mean,
            floating[:] noise_variance,
            integral[:] cov_unmasked,
            floating[:, :] block,
            integral[:] unmasked,
            integral[:] ustart,
            integral[:] uend,
            floating[:] features,
            integral[:] vstart,
            integral[:] vend,
            floating[:] f2m,
            floating[:] ct,
            floating[:] correction_term,
            integral num_features,
            ):
    cdef integral i, j, k, ip, num_unmasked, p
    cdef integral num_spikes = len(spike_indices)
    for ii in range(num_spikes):
        p = spike_indices[ii]
        for i in range(num_features):
            f2m[i] = noise_mean[i]-cluster_mean[i]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            j = unmasked[ustart[p]+i]
            k = vstart[p]+i
            f2m[j] = features[k]-cluster_mean[j]
            
        num_unmasked = len(cov_unmasked)
        for i in range(num_unmasked):
            for j in range(num_unmasked):
                block[i, j] += f2m[cov_unmasked[i]]*f2m[cov_unmasked[j]]
        
        for i in range(num_features):
            ct[i] = noise_variance[i]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            ct[unmasked[ustart[p]+i]] = correction_term[vstart[p]+i]
        
        num_unmasked = len(cov_unmasked)
        for i in range(num_unmasked):
            block[i, i] += ct[cov_unmasked[i]]


cpdef do_var_accum_mua(
            integral[:] spike_indices,
            floating[:] cluster_mean,
            floating[:] noise_mean,
            floating[:] noise_variance,
            integral[:] cov_unmasked,
            floating[:, :] block,
            integral[:] unmasked,
            integral[:] ustart,
            integral[:] uend,
            floating[:] features,
            integral[:] vstart,
            integral[:] vend,
            floating[:] f2m,
            floating[:] ct,
            floating[:] correction_term,
            integral num_features,
            ):
    cdef integral i, j, k, ip, num_unmasked, p
    cdef integral num_spikes = len(spike_indices)
    for ii in range(num_spikes):
        p = spike_indices[ii]
        for i in range(num_features):
            f2m[i] = noise_mean[i]-cluster_mean[i]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            j = unmasked[ustart[p]+i]
            k = vstart[p]+i
            f2m[j] = features[k]-cluster_mean[j]
            
        num_unmasked = len(cov_unmasked)
        for i in range(num_unmasked):
            block[i, i] += f2m[cov_unmasked[i]]*f2m[cov_unmasked[i]]
        
        for i in range(num_features):
            ct[i] = noise_variance[i]
        num_unmasked = uend[p]-ustart[p]
        for i in range(num_unmasked):
            ct[unmasked[ustart[p]+i]] = correction_term[vstart[p]+i]
        
        num_unmasked = len(cov_unmasked)
        for i in range(num_unmasked):
            block[i, i] += ct[cov_unmasked[i]]
        


"""Installation script."""


#------------------------------------------------------------------------------
# Imports
#------------------------------------------------------------------------------

import os
import os.path as op
import re
import numpy
import sys
import platform

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

from Cython.Build import cythonize

#------------------------------------------------------------------------------
# Setup
#------------------------------------------------------------------------------


# Copied from Brian 2
def has_option(name):
    try:
        sys.argv.remove('--%s' % name)
        return True
    except ValueError:
        pass
    # allow passing all cmd line options also as environment variables
    env_val = os.getenv(name.upper().replace('-', '_'), 'false').lower()
    if env_val == "true":
        return True
    return False


with_openmp = has_option('with-openmp')
no_openmp = has_option('no-openmp')

def _package_tree(pkgroot):
    path = op.dirname(__file__)
    subdirs = [op.relpath(i[0], path).replace(op.sep, '.')
               for i in os.walk(op.join(path, pkgroot))
               if '__init__.py' in i[2]]
    return subdirs


curdir = op.dirname(op.realpath(__file__))
readme = open(op.join(curdir, 'README.rst')).read()


# Find version number from `__init__.py` without executing it.
filename = op.join(curdir, 'klustakwik2/__init__.py')
with open(filename, 'r') as f:
    version = re.search(r"__version__ = '([^']+)'", f.read()).group(1)


requirements = [
]

test_requirements = [
]

extensions = cythonize('klustakwik2/numerics/cylib/*.pyx')
for ext in extensions:
    if 'e_step_cy' in ext.name:
        if no_openmp:
            continue
        if os.name=='nt': # Windows
            ext.extra_compile_args = ['/openmp']
        elif platform.system()=='Darwin' and not with_openmp: # Mac
            pass
        else:
            ext.extra_compile_args =['-fopenmp']
            ext.extra_link_args = ['-fopenmp']

setup_kwds = dict(
    name='klustakwik2',
    version=version,
    description='Clustering for high dimensional neural data',
    long_description=readme,
    url='https://github.com/kwikteam/klustakwik2',
    packages=_package_tree('klustakwik2'),
    package_dir={'klustakwik2': 'klustakwik2'},
    package_data={'klustakwik2.numerics.cylib': ['*.pyx']},
    include_package_data=True,
    ext_modules=extensions,
    include_dirs=[numpy.get_include()],
    entry_points={
        'console_scripts': [
            'kk2_legacy=klustakwik2.scripts.kk2_legacy:main',
        ],
    },
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='klustakwik2,data analysis,electrophysiology,neuroscience,clustering',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    tests_require=test_requirements
    )

try:
    setup(**setup_kwds)
except SystemExit as e:
    # try without OpenMP
    for ext in extensions:
        ext.extra_compile_args = []
        ext.extra_link_args = []
    setup(**setup_kwds)



'''
Logging module

Written as a wrapper around the Python logging module. Making it a wrapper should make it easier
to refactor it into the logging solution of phy later on.
'''

import logging

__all__ = ['logger', 'log_to_file', 'file_log_level', 'console_log_level', 'log_message',
           'log_suppress_hierarchy', 'log_suppress_name', 'log_remove_filters',
           ]

logger = logging.getLogger('klustakwik')
logger.propagate = False
logger.setLevel(logging.DEBUG)

#: Translation from string representation to number
LOG_LEVELS = {'CRITICAL': logging.CRITICAL,
              'ERROR': logging.ERROR,
              'WARNING': logging.WARNING,
              'INFO': logging.INFO,
              'DEBUG': logging.DEBUG}

FILE_HANDLER = None

def log_to_file(fname, level=logging.INFO):
    if isinstance(level, str):
        level = LOG_LEVELS[level.upper()]
    FILE_HANDLER = logging.FileHandler(fname, 'wt')
    FILE_HANDLER.setLevel(level)
    FILE_HANDLER.setFormatter(logging.Formatter('%(asctime)s %(levelname)-8s %(name)s: %(message)s'))
    logger.addHandler(FILE_HANDLER)
    
def file_log_level(level):
    if isinstance(level, str):
        level = LOG_LEVELS[level.upper()]
    FILE_HANDLER.setLevel(level)

# create console handler with a higher log level
CONSOLE_HANDLER = logging.StreamHandler()
CONSOLE_HANDLER.setLevel(logging.INFO)
CONSOLE_HANDLER.setFormatter(logging.Formatter('%(levelname)-8s %(name)s: %(message)s'))
logger.addHandler(CONSOLE_HANDLER)

def console_log_level(level):
    if isinstance(level, str):
        level = LOG_LEVELS[level.upper()]
    CONSOLE_HANDLER.setLevel(level)

def log_message(level, msg, name=None):
    '''
    Adds a log message at the specified log level.
    
    Log levels (strings) can be one of (In descending order of severity): critical, error, warning,
    info, debug.
    '''
    if isinstance(level, str):
        level = LOG_LEVELS[level.upper()]
    if name is None or name=='':
        name = 'klustakwik'
    else:
        name = 'klustakwik.'+name
    logger = logging.getLogger(name)
    logger.log(level, msg)


class HierarchyFilter(object):
    def __init__(self, name='', inclusive=True):
        self.name = name
        self.nlen = len(name)
        self.inclusive = inclusive

    def filter(self, record):
        if self.nlen == 0:
            return 0
        elif self.inclusive and self.name==record.name:
            return 0
        elif not self.inclusive and self.name==record.name:
            return 1
        elif record.name.find(self.name, 0, self.nlen) != 0:
            return 1
        return not (record.name[self.nlen] == ".")


# See brian2.utils.logger for details
class NameFilter(object):
    '''
    A class for suppressing log messages ending with a certain name.
    
    Parameters
    ----------
    name : str
        The name to suppress. See `BrianLogger.suppress_name` for details.
    '''
    
    def __init__(self, name):
        self.name = name
    
    def filter(self, record):
        '''
        Filter out all messages ending with a certain name.
        '''
        # The last part of the name
        record_name = record.name.split('.')[-1]
        return self.name != record_name


def log_suppress_hierarchy(name, inclusive=True, console=True, file=False):
    filter = HierarchyFilter(name, inclusive=inclusive)
    if console:
        CONSOLE_HANDLER.addFilter(filter)
    if file:
        FILE_HANDLER.addFilter(filter)

def log_suppress_name(name, console=True, file=False):
    filter = NameFilter(name)
    if console:
        CONSOLE_HANDLER.addFilter(filter)
    if file:
        FILE_HANDLER.addFilter(filter)

def log_remove_filters(console=True, file=True):
    if console:
        CONSOLE_HANDLER.filters = []
    if file:
        FILE_HANDLER.filters = []



from numpy import *
from klustakwik2 import *
from numpy.testing import assert_raises, assert_array_almost_equal, assert_array_equal
from nose import with_setup
from nose.tools import nottest
from numpy.random import randint, rand
from six.moves import range

from .test_mask_starts import generate_multimask_test_data

def test_partitioning():
    for _ in range(100):
        # actually we only need to generate some arbitrary data, the values don't matter much
        data = generate_multimask_test_data(10, 10, 5)
        kk = KK(data)
        clusters = [0, 1, 0, 2, 0, 2, 0, 3, 0, 3]
        kk.initialise_clusters(clusters)
        assert_array_equal(kk.num_cluster_members, [5, 1, 2, 2])
        assert_array_equal(kk.get_spikes_in_cluster(0), [0, 2, 4, 6, 8])
        assert_array_equal(kk.get_spikes_in_cluster(1), [1])
        assert_array_equal(kk.get_spikes_in_cluster(2), [3, 5])
        assert_array_equal(kk.get_spikes_in_cluster(3), [7, 9])
        assert_array_equal(kk.spikes_in_cluster, [0, 2, 4, 6, 8, 1, 3, 5, 7, 9])
        assert_array_equal(kk.spikes_in_cluster_offset, [0, 5, 6, 8, 10])
    

if __name__=='__main__':
    test_partitioning()
